from pytest import importorskip


def test_simple():
    importorskip("examples.simple")

def test_reference():
    importorskip("examples.reference")

def test_nested():
    importorskip("examples.nested")

def test_identity():
    importorskip("examples.identity")



from pytest import importorskip


def test_simple():
    importorskip("examples.generic.simple")


def test_nested():
    importorskip("examples.generic.nested")


def test_reference():
    importorskip("examples.generic.reference")


def test_identity():
    importorskip("examples.generic.identity")



from typing import Annotated, Any, Generic, TypeVar, Union

from pydantic import BaseModel, BeforeValidator, TypeAdapter, model_validator

__all__ = ("R", "Route", "Routed", "Router")

RoutePart = Union[int, str]
rp_ta = TypeAdapter(RoutePart)


def split_route(route: str) -> list[RoutePart]:
    """Split a `.`-separated string/integer subpath up into a list."""
    match route:
        case str():
            return list(map(rp_ta.validate_strings, route.split(".")))
        case list():
            return route
        case _:
            raise ValueError(f"Invalid route: {route}")


Route = Annotated[list[RoutePart], BeforeValidator(split_route)]
"""The `.`-separated string keys and integers that specify a JSON-like subpath."""


class Router(BaseModel, validate_default=True):
    """A model which should be subclassed to specify `Route` type fields).

    For example:

    ```py
    from fieldrouter import Router, Route

    class Where(Router):
        value_1: Route = "example.subpath.0"
    ```
    """


R = TypeVar("R", bound=Router)


def extract_subpath(path: Route, data: dict) -> Any:
    """Extract a subpath or else an error reporter fallback indicating where it failed."""
    for part_idx, part in enumerate(path):
        reporter = ValueError(f"Missing {part=} on {path}")
        match part:
            case str() as key:
                data = data.get(key, reporter)
            case int() as idx:
                data = (
                    data[idx]
                    if isinstance(data, list) and -len(data) <= idx < len(data)
                    else reporter
                )
        if data is reporter:
            break
    return data


class Routed(BaseModel, Generic[R]):
    """A model which should be subclassed to specify fields at the associated routes.

    When using, the router subclass (with the same field names as this model) should be
    passed as its generic type argument.

    ```py
    from fieldrouter import R, Route, Routed, Router


    class Where(Router):
        some_value: Route = "example.subpath.0"


    class What(Routed[R]):
        some_value: int


    data = {"example": {"subpath": [100]}}
    result = What[Where].model_validate(data).some_value  # 100
    ```
    """

    @model_validator(mode="before")
    def supply_routes(cls, data: dict):
        values = dict(data)  # Make a copy
        route_model = cls.__pydantic_generic_metadata__["args"][0]
        router = route_model()
        for field, route in router.model_dump().items():
            if is_identity_route := route == ["", ""]:  # noqa:F841
                values.update({field: data})
                continue
            elif has_referent_root := route[0] == "":  # noqa:F841
                # This could fail if the referent doesn't exist or the route's malformed
                referent = route[1]
                path = route[2:]
                source = values[referent]
            else:
                path = route
                source = data
            values.update({field: extract_subpath(path=path, data=source)})
        for unrouted_field in set(cls.model_fields) - set(router.model_fields):
            reporter = ValueError(f"No route for {unrouted_field}")
            values.update({unrouted_field: reporter})
        return values



from typing import Annotated, Any, TypeVar, Union

from pydantic import BaseModel, BeforeValidator, TypeAdapter, model_validator

__all__ = ("Route", "Routing", "RoutingModel")

RoutePart = Union[int, str]
rp_ta = TypeAdapter(RoutePart)


def split_route(route: str) -> list[RoutePart]:
    """Split a `.`-separated string/integer subpath up into a list."""
    match route:
        case str():
            return list(map(rp_ta.validate_strings, route.split(".")))
        case list():
            return route
        case _:
            raise ValueError(f"Invalid route: {route}")


Route = Annotated[list[RoutePart], BeforeValidator(split_route)]
"""The `.`-separated string keys and integers that specify a JSON-like subpath."""


class Via:
    routes: Route | list[Route]

    def __init__(self, routes):
        self.routes = routes

    def __repr__(self):
        return f"Via(routes={self.routes})"


T = TypeVar("T")


def Routing(tp: T, via: str | list[str]) -> Annotated[T, Via]:
    match via:
        case list():
            routes = via
        case str():
            routes = [via]
        case _ as t:
            raise TypeError(f"Expected a route string or list of strings, got {t}")
    return Annotated[tp, Via(routes=[via])]


def extract_subpath(path: Route, data: dict) -> Any:
    """Extract a subpath or else an error reporter fallback indicating where it failed."""
    if path == ["", ""]:
        return data
    for part_idx, part in enumerate(path):
        reporter = ValueError(f"Missing {part=} on {path}")
        match part:
            case str() as key:
                data = data.get(key, reporter)
            case int() as idx:
                data = (
                    data[idx]
                    if isinstance(data, list) and -len(data) <= idx < len(data)
                    else reporter
                )
        if data is reporter:
            break
    return data


class RoutingModel(BaseModel):
    """A model which should be subclassed to specify fields at the associated routes.

    ```py
    from typing import Annotated
    from fieldrouter.v2 import Route, RoutingModel


    class DataRouter(RoutingModel):
        some_value: Annotated[int, Route("example.subpath.0")]


    data = {"example": {"subpath": [100]}}
    result = DataRouter.model_validate(data).some_value  # 100
    ```
    """

    @model_validator(mode="before")
    def supply_routes(cls, data: dict):
        values = dict(data)  # Make a copy to mutate
        for field, route_meta in cls.model_fields.items():
            match route_meta.metadata:
                case [Via()]:
                    route_string = route_meta.metadata[0].routes[0]
                    route = split_route(route_string)
                    # A field reference is when the route begins with a . (e.g. ".foo")
                    # except the identity route (".") which reproduces the entire input
                    if route[0] == "" and route[1:] != [""]:
                        # This could fail if the referent doesn't exist or the route's malformed
                        if (referent := route[1]) not in values:
                            raise NameError(f"No such referent {referent}")
                        path = route[2:]
                        source = values[referent]
                    else:
                        path = route
                        source = data
                    route_data = extract_subpath(path=path, data=source)
                    values[field] = route_data
                case _:
                    pass  # Do not otherwise interfere with the data
        return values
